import Color from "./Color";
import MessageHandler from "./MessageHandler"
import Rect from "./Rect";

export default class CPDFAnnotation {
  messageHandler: MessageHandler;
  ptr: number;
  pageNumber: number;
  hidden: boolean;
  type: string;
  status: number = 0

  private _color: Color;
  private _rect: Rect;
  private _annotPtr: number;
  private _content: string;
  private _author: string | undefined;
  private _title: string | undefined;
  private _creationDate: Date | undefined;
  private _modifyDate: Date | undefined;
  private _transparency: number;

  constructor (options: {
    messageHandler: MessageHandler;
    annotPtr: number;
    ptr: number;
    rect: {
      left: number;
      top: number;
      right: number;
      bottom: number;
    },
    color: {
      R: number;
      G: number;
      B: number;
    };
    pageNumber: number;
    hidden: any;
    type: any;
    content: string;
    transparency?: number;
    createDate?: Date;
    modifyDate?: Date
  }) {
    this.messageHandler = options.messageHandler;
    this.ptr = options.ptr;
    this.pageNumber = options.pageNumber;
    this._color = new Color(options.color.R, options.color.G, options.color.B)
    this._rect = new Rect({
      left: options.rect.left,
      top: options.rect.top,
      right: options.rect.right,
      bottom: options.rect.bottom
    })
    this.hidden = options.hidden;
    this.type = options.type;
    this._content = options.content
    this._annotPtr = options.annotPtr
    this._creationDate = options.createDate
    this._modifyDate = options.modifyDate
    this._transparency = options.transparency || 1
  }

  getPageNumber () {
    return this.pageNumber;
  }

  getAnnotType () {
    return this.type;
  }

  get rect () {
    return this._rect;
  }

  async setRect (rect: any) {
    const { left, right, bottom, top } = rect
    if (rect.left === this.rect.left && rect.top === this.rect.top && rect.right === this.rect.right && rect.bottom === this.rect.bottom) return

    await this.messageHandler.sendWithPromise('SetAnnotRect', {
      ptr: this.ptr,
      annotPtr: this._annotPtr,
      rect: {
        left,
        right,
        bottom,
        top
      }
    })
    this.rect.setRect({
      left,
      right,
      bottom,
      top
    })
  }
  
  get color () {
    return this._color.getColor()
  }

  async setColor (value: string) {
    const { R, G, B } = this._color.formatColor(value)
    await this.messageHandler.sendWithPromise('SetAnnotRGBColor', {
      annotPtr: this._annotPtr,
      color: {
        R, G, B
      }
    })
  }


  get content() {
    return this._content
  }

  async setContent(value: string) {
    if (value === this._content) return
    await this.messageHandler.sendWithPromise('SetAnnotContent', {
      annotPtr: this._annotPtr,
      content: value
    })
    this._content = value
  }

  get author () {
    return this._author
  }

  get title () {
    return this._title;
  }

  get creationDate () {
    return this._creationDate;
  }

  get modifyDate () {
    return this._modifyDate;
  }

  get transparency () {
    return this._transparency;
  }

  async setTransparency (value: number) {
    if (value === this._transparency) return

    await this.messageHandler.sendWithPromise('SetAnnotTransparency', {
      annotPtr: this._annotPtr,
      transparency: value
    })
    this._transparency = value
  }

}

