import { AnnotationType } from './type/index'
import MessageHandler from "./MessageHandler"
import { applyStyle, getDevicePixelRatio } from "./utils"
import TextAnnotation from './CPDFTextAnnotation'

type Page = {
  doc: number;
  viewerContainer: HTMLDivElement,
  pageCount: number;
  pageNumber: number;
  messageHandler: MessageHandler;
  zoom: number;
  rotation: number;
}

type pageSize = {
  width: number;
  height: number;
  page: number;
}

enum State {
  init = 0,
  loading = 1,
  loaded = 2
}

export default class CPDFPage {
  viewerContainer: HTMLDivElement
  doc: number
  pageContainer: HTMLDivElement | null = null
  pageCount: number
  pageNumber: number
  pageSize: pageSize | null = null
  zoom: number
  rotation: number
  messageHandler: MessageHandler
  annotations: any = []
  annotationCanvas: HTMLCanvasElement | null = null
  state: State = 0
  ratio: number = 1
  annotationsType: any

  constructor({
    doc,
    viewerContainer,
    pageCount,
    pageNumber,
    messageHandler,
    zoom,
    rotation
  }: Page) {
    this.viewerContainer = viewerContainer
    this.doc = doc
    this.pageCount = pageCount
    this.pageNumber = pageNumber
    this.messageHandler = messageHandler
    this.zoom = zoom
    this.rotation = rotation

    this.ratio = getDevicePixelRatio()

    this.annotationsType = {
      text: TextAnnotation
    }
  }

  get width() {
    return this.pageSize?.width! * this.zoom
  }

  get height() {
    return this.pageSize?.height! * this.zoom
  }

  get page() {
    return this.pageSize?.page
  }

  async init() {
    const pageSize = await this.getPageSize(this.pageNumber)
    this.pageSize = pageSize as pageSize
    await this.getAnnotations()
    this.render()
  }

  async getPageSize(pageIndex: number) {
    const pageSize = await this.messageHandler?.sendWithPromise("GetPageSize", {
      doc: this.doc,
      pageIndex
    })
    return pageSize
  }

  render() {
    const pageContainer = document.createElement('div')
    this.pageContainer = pageContainer
    applyStyle(pageContainer, {
      position: 'relative',
      width: this.width,
      height: this.height,
      margin: '20px auto',
      backgroundColor: '#fff'
    })

    this.viewerContainer.appendChild(pageContainer)

    this.state = 1
  }

  async draw() {
    const data: any = await this.messageHandler?.sendWithPromise("PushRenderTask", {
      page: this.page,
      doc: this.doc,
      width: this.width,
      height: this.height,
      zoom: this.zoom
    })
    const { imageArray }: {
      imageArray: Uint8Array,
    } = data

    const canvas = document.createElement('canvas')
    canvas.setAttribute('width', this.width.toString())
    canvas.setAttribute('height', this.height.toString())
    applyStyle(canvas, {
      position: 'absolute',
      width: '100%',
      height: '100%'
    })
    this.pageContainer!.appendChild(canvas)


    let drawContext = canvas.getContext("2d")!
    let imageData = new ImageData(this.width, this.height);
    imageData.data.set(imageArray)
    drawContext.putImageData(imageData, 0, 0)

    const annotationCanvas = document.createElement('canvas')
    annotationCanvas.setAttribute('width', this.width.toString())
    annotationCanvas.setAttribute('height', this.height.toString())
    this.annotationCanvas = annotationCanvas
    applyStyle(annotationCanvas, {
      width: this.width,
      height: this.height,
      zIndex: 20,
      position: 'absolute'
    })
    this.pageContainer!.appendChild(annotationCanvas)

    this.state = 2
  }

  reset() {
    this.pageContainer!.innerHTML = ''
    this.state = 1
  }

  async getAnnotations() {
    const data = await this.messageHandler?.sendWithPromise('GetAnnot', {
      pageIndex: this.pageNumber,
      page: this.page
    }) as any
    for (let i = 0; i < data.annotCount; i++) {
      const annotation = data.annotations[i]
      const type = AnnotationType[annotation.type]
      annotation.pageNumber = this.pageNumber
      annotation.type = type
      const attr = await this.getAnnotationAttr(type, annotation.annotPtr)
      Object.assign(annotation, attr)
      if (this.annotationsType[type]) {
        this.annotations.push(new this.annotationsType[type](annotation))
      }
      console.log('annotation', annotation)
    }
    console.log(this.annotations)
  }

  async getAnnotationAttr(type: string, annotPtr: number) {
    switch (type) {
      case 'text':
        return await this.messageHandler?.sendWithPromise('GetAnnotText', {
          annotPtr,
          page: this.page
        })
        break
      case 'link':
        return await this.messageHandler?.sendWithPromise('GetAnnotLink', {
          annotPtr,
          page: this.page
        })
        break
      case 'freeText':
        
        break
      case 'line':
        
        break
      case 'square':
        
        break
      case 'circle':
        
        break
      case 'polygon':
        
        break
      case 'polyline':
        
        break
      case 'highlight':
        
        break
      case 'underline':
        
        break
      case 'squiggly':
        
        break
      case 'strikeout':
        
        break
      case 'stamp':
        
        break
      case 'ink':
        
        break
      case 'widge':
        
        break
    }
  }
}
