// @ts-nocheck
class canvasManager {
  public useCache: boolean = true
  public cache: any = {}

  constructor() {
  }

  setCanvasSize(canvas: HTMLCanvasElement, width: number, height: number) {
    canvas.setAttribute("width", width.toString());
    canvas.setAttribute("height", height.toString());
    canvas.style.width = width + "px";
    canvas.style.height = height + "px";
  }

  createCanvas(width: number, height: number) {
    var canvas = document.createElement("canvas");
    this.setCanvasSize(canvas, width, height);
    return canvas;
  }

  copyCanvas(source: HTMLCanvasElement, destination: HTMLCanvasElement) {
    var contextSource = source.getContext("2d");
    var contextDestination = destination.getContext("2d");
    contextDestination?.clearRect(0, 0, destination.width, destination.height);
    contextDestination?.drawImage(source, 0, 0);
  }

  getImageDataAlpha(canvas) {
    var context = canvas.getContext("2d");
    return context.getImageData(0, 0, canvas.width, canvas.height).data[3];
  }

  flattenCanvas(canvas) {
    var imageAlpha = this.getImageDataAlpha(canvas);
    if (imageAlpha === 0) {
      canvas.width = 0;
      canvas.height = 0;
    }
    return canvas;
  }

  rotateCanvas(canvas, degrees) {
    var tmpCanvas = this.createCanvas(canvas.height, canvas.width);
    var tmpContext = tmpCanvas.getContext("2d");
    tmpContext.translate(canvas.height / 2, canvas.width / 2);
    tmpContext.rotate((Math.PI / 180) * degrees);
    tmpContext.drawImage(canvas, -canvas.width / 2, -canvas.height / 2);
    this.copyCanvas(tmpCanvas, canvas);
  }

  cropCanvas(canvas, x, y, width, height) {
    var tmpCanvas = this.createCanvas(width, height);
    var tmpContext = tmpCanvas.getContext("2d");
    tmpContext.drawImage(canvas, x, y, width, height, 0, 0, width, height);
    this.setCanvasSize(canvas, width, height);
    this.copyCanvas(tmpCanvas, canvas);
  }

  drawImage(source, destination, transform) {
    var context = destination.getContext("2d");
    context.clearRect(0, 0, destination.width, destination.height);
    context.transform(transform.m_a, transform.m_b, transform.m_c, transform.m_d, transform.m_h, transform.m_v);
    context.drawImage(source, 0, 0);
  }
  applyTransform(canvas, transform, width, height, position, rotation) {
    if (rotation && !isNaN(rotation)) {
      this.rotateCanvas(canvas, rotation);
    }
    canvas.style.left = position.x + "px";
    canvas.style.top = position.y + "px";
    canvas.style.transform = "";
    if (width !== parseFloat(canvas.style.width) || height !== parseFloat(canvas.style.height)) {
      canvas.width = width;
      canvas.height = height;
      canvas.style.width = width + "px";
      canvas.style.height = height + "px";
    }
    var context = canvas.getContext("2d");
    context.setTransform(transform.m_a, transform.m_b, transform.m_c, transform.m_d, transform.m_h, transform.m_v);
    return canvas;
  }

  createAuxiliaryCanvas(index, width, height, scale, rotation) {
    if (!this.cache[index] || !this.cache[index][width]) {
      var canvas = this.createCanvas(width * scale, height * scale);
      this.flattenCanvas(canvas);
      if (!this.cache[index]) {
        this.cache[index] = {};
      }
      this.cache[index][width] = canvas;
    }
    var cachedCanvas = this.cache[index][width];
    var auxiliaryCanvas = this.createCanvas(width * scale, height * scale);
    this.copyCanvas(cachedCanvas, auxiliaryCanvas);
    this.applyTransform(auxiliaryCanvas, null, width * scale, height * scale, null, rotation);
    return auxiliaryCanvas;
  }

  removeAuxiliaryCanvas(index, width) {
    if (this.cache[index] && this.cache[index][width]) {
      var canvas = this.cache[index][width];
      canvas.width = 0;
      canvas.height = 0;
      delete this.cache[index][width];
    }
  };

  clearCanvas(canvas) {
    var context = canvas.getContext("2d");
    context.clearRect(0, 0, canvas.width, canvas.height);
  };

  getPageContainer(index, container) {
    return container.querySelector("#pageContainer" + index);
  }

  getWidgetContainer(index, container) {
    return container.querySelector("#pageWidgetContainer" + index);
  }

  delayedRemove(canvas) {
    setTimeout(function () {
      var host = canvas.getRootNode();
      var inShadowRoot = host instanceof ShadowRoot;
      var hostExists = document.body.contains(host);
      var isShadowRoot = inShadowRoot && hostExists;
      if (!document.body.contains(canvas) && !isShadowRoot) {
        canvas.width = 0;
        canvas.height = 0;
      }
    },
      5000);
  }
}
