import { PromiseExt, getVisibleElements } from './utils';
import Annotations from './CPDFAnnotations';

function getId(): string {
  function random() {
    return (65536 * (1 + Math.random()) | 0).toString(16).substring(1)
  }
  return random() + random() + "-" + random() + "-" + random() + "-" + random() + "-" + random() + random() + random()
}
import EventManager from "./EventManager"
import MessageHandler from "./MessageHandler"
import Page from './CPDFPage'

enum ScrollMode {
  VERTICAL = 0, // Default value.
  HORIZONTAL = 1
}

export default class DocumentViewer extends EventManager {
  id: string
  cursor: string
  defaults: {
    zoom: 1,
    rotation: 0
  }
  Annotations: Annotations | null = null
  container: HTMLDivElement | null = null
  viewerContainer: HTMLDivElement | null = null
  doc: number | null = null
  worker: null | PDFWorker = null
  messageHandler: MessageHandler | null = null
  pageCount: number = 0
  pages: Page[] = []
  scrollMode: ScrollMode = 0
  cacheBuf: Set<Page> | null = new Set()
  private _currentPageNumber: number = 0

  constructor() {
    super()

    this.id = getId()
    this.cursor = "default"

    this.defaults = {
      zoom: 1,
      rotation: 0
    }

    this.initDocument()
  }

  initDocument() {
    const worker = new PDFWorker()

    this.worker = worker
    const messageHandler = worker.messageHandler
    this.messageHandler = messageHandler

    this.setupMessageHandler()

    this.Annotations = new Annotations({
      doc: this.doc!,
      pages: this.pages,
      messageHandler: this.messageHandler
    })
  }

  setupMessageHandler() {
    const messageHandler = this.messageHandler 

    messageHandler?.on('initDoc', ({ doc }) => {
      this.doc = doc
    })
  }

  async loadDocument(file: File, options: {
    container: HTMLDivElement,
    viewerContainer: HTMLDivElement
  }) {
    this.container = options.container
    this.viewerContainer = options.viewerContainer
    const messageHandler = this.messageHandler 
    await new Promise<void>((resolve, reject) => {
      if (file instanceof Uint8Array) {
        let buffer = new Uint8Array(file)
        messageHandler?.sendWithPromise('LoadFile', {
          buffer
        })
        messageHandler?.sendWithPromise('LoadDocumentByStream', {
          doc: this.doc,
          fileId: 0,
          length: buffer.length
        })
        resolve()
      } else {
        let reader = new FileReader()
        reader.onload = async ({ target }) => {
          try {
            
            let buffer = new Uint8Array(target?.result as ArrayBuffer)

            await messageHandler?.sendWithPromise('LoadFile', {
              buffer
            })

            await messageHandler?.sendWithPromise('LoadDocumentByStream', {
              doc: this.doc,
              fileId: 0,
              length: buffer.length
            })
            resolve()
          } catch (error) {
            reject(error)
          }
        }

        reader.readAsArrayBuffer(file)
      }
    })

    const pageCount = await this.getPageCount() as number
    this.pageCount = pageCount
    for (let pageNumber = 0; pageNumber < pageCount; pageNumber++) {
      const page = new Page({
        doc: this.doc!,
        viewerContainer: this.viewerContainer,
        pageCount,
        pageNumber,
        zoom: this.defaults.zoom,
        rotation: this.defaults.rotation,
        messageHandler: this.messageHandler!
      })

      await page.init()
      this.pages.push(page)
    }
    const visible = this._getVisiblePages()
    console.log(visible)

    const visiblePages = visible.views,
      numVisiblePages = visiblePages.length;

    if (numVisiblePages === 0) {
      return;
    }

    if (visible.ids) {
      const buf = this.cacheBuf!
      for (let i = 0; i < visible.views.length; i++) {
        const view = visible.views[i].view
        if (!buf.has(view)) {
          buf.add(view)
        }
      }
      let i = 0;
      const ii = buf.size;
      for (const view of buf) {
        if (visible.ids.has(view.pageNumber)) {
          buf.delete(view) // Move the view to the "end" of the buffer.
          buf.add(view)
          if (view.state !== 2) {
            await view.draw()
            this.Annotations?.addAnnotation(view.annotations)
          }
        } else {
          view.reset()
          buf.delete(view)
        }
        if (++i >= ii) {
          break;
        }
      }
    }

    const currentId = this._currentPageNumber;
    let stillFullyVisible = false;

    for (const page of visiblePages) {
      if (page.percent < 100) {
        break;
      }
      if (page.id === currentId) {
        stillFullyVisible = true;
        break;
      }
    }
    this._setCurrentPageNumber(
      stillFullyVisible ? currentId : visiblePages[0].id
    )
    
  }

  _setCurrentPageNumber(val: number) {
    if (this._currentPageNumber === val) {
      return true;
    }
    this._currentPageNumber = val
  }

  async getPageCount() {
    const pageNum = await this.messageHandler?.sendWithPromise('GetPageCount', {
      doc: this.doc
    })

    return pageNum
  }

  _getVisiblePages() {
    const views = this.pages
    const horizontal = this.scrollMode === 1
    return getVisibleElements({
      scrollEl: this.container!,
      views,
      horizontal
    })
  }
}

class PDFWorker {
  destroyed = false
  private _promiseCapability = new PromiseExt()
  private _messageHandler: MessageHandler | null = null
  private _webWorker: Worker | null = null

  constructor() {
    this._initialize()
  }

  private _initialize() {
    try {
      const workerSrc: string = "./lib/pdfviewer.worker.js"
      const worker: Worker = new Worker(workerSrc)
      const messageHandler = new MessageHandler('main', 'worker', worker)
      this._messageHandler = messageHandler

      const terminateEarly = () => {
        worker.removeEventListener("error", onWorkerError)
        messageHandler.destroy()
        worker.terminate()
        if (this.destroyed) {
          this._promiseCapability.reject(new Error("Worker was destroyed"))
        } else {
          console.error("The worker has been disabled.")
        }
      }

      const onWorkerError = () => {
        if (!this._webWorker) {
          // Worker failed to initialize due to an error. Clean up and fall
          // back to the fake worker.
          terminateEarly();
        }
      }

      worker.addEventListener("error", onWorkerError)

      messageHandler.on("test", (data: MessageEvent) => {
        worker.removeEventListener("error", onWorkerError)
        if (this.destroyed) {
          terminateEarly();
          return; // worker was destroyed
        }
        if (data) {
          this._messageHandler = messageHandler
          this._webWorker = worker

          this._promiseCapability.resolve(true);
        }
      })

      messageHandler.on("ready", data => {
        worker.removeEventListener("error", onWorkerError);
        if (this.destroyed) {
          terminateEarly();
          return; // worker was destroyed
        }
        try {
          sendTest();
          
          messageHandler.send("Ready", true)
        } catch {
          console.error("The worker has been disabled.")
        }
      })

      const sendTest = () => {
        const testObj = new Uint8Array();
        // Ensure that we can use `postMessage` transfers.
        messageHandler.send("test", testObj);
      }

      sendTest()
    } catch {
      console.log("The worker has been disabled.")
    }
  }

  get promise() {
    return this._promiseCapability.promise;
  }

  get messageHandler() {
    return this._messageHandler;
  }

  destroy() {
    this.destroyed = true
    if (this._webWorker) {
      // We need to terminate only web worker created resource.
      this._webWorker.terminate()
      this._webWorker = null
    }
    if (this._messageHandler) {
      this._messageHandler.destroy()
      this._messageHandler = null
    }
  }
}
