import CPDFAnnotations from "./CPDFAnnotations";
class PromiseExt {
  constructor() {
    this.promise = new Promise((resolve, reject) => {
      this.resolve = (data) => {
        resolve(data)
      }
      this.reject = (data) => {
        reject(data)
      }
    })
  }
}

class DocList {
  constructor() {
    this.docId = -1
    this.docArray = []
  }

  increaseId() {
    this.docId++
    return this.docId
  }
}

let docHandler = new DocList()

function MessageHandler(evt) {
  const { uid, msgId, cmd, data } = evt.data
  if (docHandler.docArray[uid] != undefined) {
    let doc = docHandler.docArray[uid]

    if (cmd == "taskFinished") {
      doc.EmptyRenderTask()
      return;
    }

    if (doc.messagehandler.messages[msgId] != undefined) {
      if (cmd == "PushRenderTask") {
        doc.appendRenderImage(data)
      }
      else {
        doc.messagehandler.messages[msgId].resolve(data)
        Reflect.deleteProperty(doc.messagehandler.messages, msgId)
      }
    }
  }
}

class MessageLoop {
  constructor() {
    this.msgId = 0
    this.messages = []
    this.workJs = new Worker("./lib/pdfviewer.worker.js")
    this.workJs.addEventListener("message", MessageHandler)
  }
  SendMessage(uid, cmd, data) {
    let promiseExt = new PromiseExt()
    this.messages[this.msgId] = promiseExt
    this.workJs.postMessage({
      uid,
      msgId: this.msgId,
      cmd,
      data
    })
    this.msgId++
    return promiseExt.promise
  }

  Terminate() {
    this.workJs.terminate()
  }
}

export class CPDFDocumentProxy {
  constructor() {
    this.runtime = false
    this.messagehandler = new MessageLoop()
    this.docId = docHandler.increaseId()
    this.renderTasks = []
    this.taskStatue = "empty"
    this.pages = []
    this.edits = []
    docHandler.docArray[this.docId] = this
  }

  appendRenderImage(data) {

    const { imageArray, pageIndex, rect } = data

    let pageDiv = document.getElementById("pageArea_" + pageIndex)
    if (!pageDiv) {
      return
    }

    let canvas = pageDiv.querySelector("canvas")
    if (!canvas) {
      canvas = document.createElement("canvas")
      canvas.style.width = "100%"
      canvas.style.height = "100%"
      pageDiv.appendChild(canvas)
    }
    canvas.width = rect.width
    canvas.height = rect.height


    let drawContext = canvas.getContext("2d")
    let imageData = new ImageData(rect.width, rect.height);
    imageData.data.set(imageArray)
    drawContext.putImageData(imageData, 0, 0)
  }

  async InitDocument() {
    let promise = this.messagehandler.SendMessage(this.docId, "initDoc")
    await promise.then((doc) => {
      this.doc = doc
    })
  }

  ClearDocument() {
    this.messagehandler.SendMessage(this.docId, "ClearDocument")
  }

  async LoadDocumentByStream(fileId, len, password) {
    await this.messagehandler.SendMessage(this.docId, "LoadDocumentByStream", {
      doc: this.doc,
      fileId,
      len,
      password
    })
  }

  async GetPageCount() {
    let pageNum = 0;
    await this.messagehandler.SendMessage(this.docId, "GetPageCount", {
      doc: this.doc
    }).then((data) => {
      pageNum = data
    })

    return pageNum
  }

  async GetPageSize(pageIndex) {
    let pageSize = undefined
    await this.messagehandler.SendMessage(this.docId, "GetPageSize", {
      doc: this.doc,
      pageIndex
    }).then((data) => {
      pageSize = data
    })
    return pageSize
  }

  async InitPage(pageIndex) {

    if (!this.pages[pageIndex]) {
      this.pages[pageIndex] = await this.messagehandler.SendMessage(this.docId, "InitPage", {
        doc: this.doc,
        pageIndex
      })
    }
    return this.pages[pageIndex]
  }

  async InsertPage(pageIndex, width, height, mode) {
    this.pages = []
    let promise = this.messagehandler.SendMessage(this.docId, "InsertPage",
      {
        doc: this.doc,
        pageIndex,
        width,
        height,
        mode
      })
    let result = false
    await promise.then((flag) => {
      result = flag
    })
    return result
  }

  async LoadFile(buf) {
    await this.messagehandler.SendMessage(this.docId, "LoadFile", buf)
  }

  async SaveDocument() {
    let dataArray = []
    await this.messagehandler.SendMessage(this.docId, "SaveDocumentByStream", {
      doc: this.doc,
      type: 2
    }).then((data) => {
      dataArray = data
    })
    return dataArray
  }

  async CreateAnnot() {
    let pagePtr = await this.InitPage(0)
    let annotPtr = await this.messagehandler.SendMessage(this.docId, "CreateAnnot",
      {
        doc: this.doc,
        page: pagePtr,
        type: 6
      })

    await this.messagehandler.SendMessage(this.docId, "SetAnnotParam",
      {
        annot: annotPtr,
        page: pagePtr
      })
    return annotPtr
  }

  async InitEditPage(pageIndex) {
    if (!this.edits[pageIndex]) {
      let pagePtr = this.pages[pageIndex]
      if (!pagePtr) {
        pagePtr = await this.InitPage(pageIndex)
      }

      this.edits[pageIndex] = await this.messagehandler.SendMessage(this.docId, "InitEditPage", pagePtr)
    }

    return this.edits[pageIndex]
  }

  async BeginEdit(editPage) {
    await this.messagehandler.SendMessage(this.docId, "BeginEdit", editPage)
  }

  async EndEdit(editPage) {
    await this.messagehandler.SendMessage(this.docId, "EndEdit", editPage)
  }

  async GetEditArea(editPage, index) {
    return await this.messagehandler.SendMessage(this.docId, "GetEditArea", { editPage, index })
  }

  async InsertText(areaPtr, text) {
    return await this.messagehandler.SendMessage(this.docId, "InsertText", { areaPtr, text })
  }

  async IsTextArea(areaPtr) {
    return await this.messagehandler.SendMessage(this.docId, "IsTextArea", areaPtr)
  }

  async GetEditAreaCount(editPage) {
    return await this.messagehandler.SendMessage(this.docId, "GetEditAreaCount", editPage)
  }

  async MovePage(from, to) {
    return await this.messagehandler.SendMessage(this.docId, "MovePage", { doc: this.doc, from, to })
  }

  // test
  async RotatePage(page, rotation) {
    this.CPDFAnnotations = new CPDFAnnotations({
      docId: this.docId,
      doc: this.doc,
      pages: this.pages,
      messagehandler: this.messagehandler
    })
    const annot = await this.CPDFAnnotations.addAnnotations({
      type: 6
    })
    // this.CPDFAnnotations.removeAnnot(annot)





    // return await this.messagehandler.SendMessage(this.docId,"RotatePage",{page,rotation})
  }

  async RemovePage(pageIndex) {
    return await this.messagehandler.SendMessage(this.docId, "RemovePage", { doc: this.doc, pageIndex })
  }

  async RemoveAnnot(annotPtr) {
    return await this.messagehandler.SendMessage(this.docId, "RemoveAnnot", annotPtr)
  }

  async SetCharsFontRederingMode(data) {
    return await this.messagehandler.SendMessage(this.docId, "SetCharsFontRederingMode", data)
  }

  async GetBeginAndEndCharPlace(areaPtr) {
    return await this.messagehandler.SendMessage(this.docId, "GetBeginAndEndCharPlace", areaPtr)
  }

  async GetTextByRange(data) {
    return await this.messagehandler.SendMessage(this.docId, "GetTextByRange", data)
  }

  async SetCharsFontSize(data) {
    return await this.messagehandler.SendMessage(this.docId, "SetCharsFontSize", data)
  }

  async SetCharsFontColor(data) {
    return await this.messagehandler.SendMessage(this.docId, "SetCharsFontColor", data)
  }

  async DeleteChar(data) {
    return await this.messagehandler.SendMessage(this.docId, "DeleteChar", data)
  }

  async CreateWidget() {
    let pagePtr = await this.InitPage(0)
    let annotPtr = await this.messagehandler.SendMessage(this.docId, "CreateWidget",
      {
        doc: this.doc,
        page: pagePtr,
        type: 1
      })

    await this.messagehandler.SendMessage(this.docId, "SetWidgetParam",
      {
        annot: annotPtr,
        page: pagePtr
      })
    return annotPtr
  }

  Close() {
    this.messagehandler.Terminate()
  }

  CacheRenderTask(visibleRange, zoom) {
    if (visibleRange.length > 0) {
      if (this.taskStatue == "empty") {
        this.renderTasks = []
        let pagesInfo = []

        for (let i = 0; i < visibleRange.length; i++) {
          let pageIndex = visibleRange[i] - 1
          let pageInfo = this.viewer.pagesInfo[pageIndex]
          if (pageInfo.isLoad == false) {
            pagesInfo.push({
              height: pageInfo.height * zoom,
              width: pageInfo.width * zoom,
              pageIndex: pageInfo.pageIndex
            })
            this.viewer.pagesInfo[pageIndex].isLoad = true
          }
        }

        if (pagesInfo.length > 0) {
          this.taskStatue = "busy"
          let pageText = ""
          pagesInfo.forEach(element => {
            pageText = pageText + " " + element.pageIndex
          });

          this.messagehandler.SendMessage(this.docId, "PushRenderTask",
            {
              doc: this.doc,
              pagesInfo,
              zoom
            })
        }
      }
      else {
        this.renderTasks.push({ visibleRange, zoom })
      }
    }
  }

  EmptyRenderTask() {
    this.taskStatue = "empty"
    if (this.renderTasks.length > 0) {
      let cacheData = this.renderTasks.pop()
      const { visibleRange, zoom } = cacheData
      this.renderTasks = []
      this.CacheRenderTask(visibleRange, zoom)
    }
  }
}
