import Page from './CPDFPage'

export class PromiseExt<T> {
  promise: Promise<T>;
  resolve!: (data: T) => void;
  reject!: (reason?: any) => void;

  constructor() {
    this.promise = new Promise<T>((resolve, reject) => {
      this.resolve = (data) => {
        resolve(data);
      };
      this.reject = (reason) => {
        reject(reason);
      };
    });
  }
}

const hyphenPattern = /-([a-z])/g

function convertToCamelCase(input: string): string {
  return input.replace(hyphenPattern, (match, letter) => letter.toUpperCase());
}

type Styles = {
  [key: string]: number | string;
};

const startsWithLowerCase = /^[a-z]/
const isMeasurementProperty = /^(?:Border(?:Top|Right|Bottom|Left)?(?:Width|)|(?:Margin|Padding)?(?:Top|Right|Bottom|Left)?|(?:Min|Max)?(?:Width|Height))$/

export function applyStyle(element: HTMLElement, styles: Styles): void {
  for (let property in styles) {
    let value = styles[property];
    property = convertToCamelCase(property);
    const cssStyle = element.style;

    if (typeof value === "number") {
      value = startsWithLowerCase.test(property) && isMeasurementProperty.test(property[0].toUpperCase() + property.slice(1)) ? value + "px" : `${value}`;
    }

    (cssStyle as any)[property as keyof CSSStyleDeclaration] = value;
  }
}

function binarySearchFirstItem(items: Page[], condition: ElementPosition, start = 0) {
  let minIndex = start;
  let maxIndex = items.length - 1;

  if (maxIndex < 0 || !condition(items[maxIndex])) {
    return items.length;
  }
  if (condition(items[minIndex])) {
    return minIndex;
  }

  while (minIndex < maxIndex) {
    const currentIndex = (minIndex + maxIndex) >> 1;
    const currentItem = items[currentIndex];
    if (condition(currentItem)) {
      maxIndex = currentIndex;
    } else {
      minIndex = currentIndex + 1;
    }
  }
  return minIndex; /* === maxIndex */
}

function backtrackBeforeAllVisibleElements(index: number, views: Page[], top: number) {
  if (index < 2) {
    return index;
  }

  let elt = views[index].pageContainer!;
  let pageTop = elt.offsetTop + elt.clientTop;

  if (pageTop >= top) {
    elt = views[index - 1].pageContainer!;
    pageTop = elt.offsetTop + elt.clientTop;
  }

  for (let i = index - 2; i >= 0; --i) {
    elt = views[i].pageContainer!;
    if (elt.offsetTop + elt.clientTop + elt.clientHeight <= pageTop) {
      // We have reached the previous row, so stop now.
      // This loop is expected to terminate relatively quickly because the
      // number of pages per row is expected to be small.
      break;
    }
    index = i;
  }
  return index;
}

type ElementPosition = (view: Page) => boolean

export function getVisibleElements ({
  scrollEl,
  views,
  horizontal = false,
}: {
  scrollEl: HTMLDivElement,
  views: Page[],
  horizontal: boolean,
}) {
  const top = scrollEl.scrollTop,
    bottom = top + scrollEl.clientHeight;
  const left = scrollEl.scrollLeft,
    right = left + scrollEl.clientWidth;

  function isElementBottomAfterViewTop(view: Page) {
    const element = view.pageContainer!;
    const elementBottom =
      element.offsetTop + element.clientTop + element.clientHeight;
    return elementBottom > top;
  }
  function isElementNextAfterViewHorizontally(view: Page) {
    const element = view.pageContainer!;
    const elementLeft = element.offsetLeft + element.clientLeft;
    const elementRight = elementLeft + element.clientWidth;
    return elementRight > left;
  }

  const visible = [],
    ids = new Set(),
    numViews = views.length;
  let firstVisibleElementInd = binarySearchFirstItem(
    views,
    horizontal
      ? isElementNextAfterViewHorizontally
      : isElementBottomAfterViewTop
  );

  if (
    firstVisibleElementInd > 0 &&
    firstVisibleElementInd < numViews &&
    !horizontal
  ) {
    firstVisibleElementInd = backtrackBeforeAllVisibleElements(
      firstVisibleElementInd,
      views,
      top
    );
  }

  let lastEdge = horizontal ? right : -1;

  for (let i = firstVisibleElementInd; i < numViews; i++) {
    const view = views[i],
      element = view.pageContainer!;
    const currentWidth = element.offsetLeft + element.clientLeft;
    const currentHeight = element.offsetTop + element.clientTop;
    const viewWidth = element.clientWidth,
      viewHeight = element.clientHeight;
    const viewRight = currentWidth + viewWidth;
    const viewBottom = currentHeight + viewHeight;

    if (lastEdge === -1) {
      if (viewBottom >= bottom) {
        lastEdge = viewBottom;
      }
    } else if ((horizontal ? currentWidth : currentHeight) > lastEdge) {
      break;
    }

    if (
      viewBottom <= top ||
      currentHeight >= bottom ||
      viewRight <= left ||
      currentWidth >= right
    ) {
      continue;
    }

    const hiddenHeight =
      Math.max(0, top - currentHeight) + Math.max(0, viewBottom - bottom);
    const hiddenWidth =
      Math.max(0, left - currentWidth) + Math.max(0, viewRight - right);

    const fractionHeight = (viewHeight - hiddenHeight) / viewHeight,
      fractionWidth = (viewWidth - hiddenWidth) / viewWidth;
    const percent = (fractionHeight * fractionWidth * 100) | 0;

    visible.push({
      id: view.pageNumber,
      x: currentWidth,
      y: currentHeight,
      view,
      percent,
      widthPercent: (fractionWidth * 100) | 0,
    });
    ids.add(view.pageNumber);
  }

  const first = visible[0],
    last = visible.at(-1);

  visible.sort(function (a, b) {
    const pc = a.percent - b.percent;
    if (Math.abs(pc) > 0.001) {
      return -pc;
    }
    return a.id - b.id;
  });
  return { first, last, views: visible, ids };
}

export function convertToPDFTimestamp(date: Date): string {
  const timezoneOffset = date.getTimezoneOffset()
  const offsetSign = timezoneOffset < 0 ? '+' : '-'
  const offsetHours = ('00' + Math.abs(timezoneOffset / 60)).slice(-2)
  const offsetMinutes = ('00' + Math.abs(timezoneOffset % 60)).slice(-2)
  const sep = `${offsetSign}${offsetHours}'${offsetMinutes}`
  const year = date.getFullYear()
  const month = ('00' + (date.getMonth() + 1)).slice(-2)
  const day = ('00' + date.getDate()).slice(-2)
  const hours = ('00' + date.getHours()).slice(-2)
  const minutes = ('00' + date.getMinutes()).slice(-2)
  const seconds = ('00' + date.getSeconds()).slice(-2)
  const timestamp = `D:${year}${month}${day}${hours}${minutes}${seconds}${sep}`

  return timestamp
}

export function parseAdobePDFTimestamp(timestamp: string): Date {
  // 使用正则表达式匹配时间戳格式
  const regex = /D:(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})([-+]\d{2}'\d{2}')/;
  const match = timestamp.match(regex);

  if (match) {
    const [, year, month, day, hour, minute, second, offset] = match;
    const offsetHours = parseInt(offset.slice(1, 3));
    const offsetMinutes = parseInt(offset.slice(4, 6));

    // 构建 Date 对象，注意月份要减 1
    const date = new Date(
      Date.UTC(
        parseInt(year),
        parseInt(month) - 1,
        parseInt(day),
        parseInt(hour),
        parseInt(minute),
        parseInt(second)
      )
    );

    // 考虑时区偏移
    date.setHours(date.getUTCHours() - offsetHours);
    date.setMinutes(date.getUTCMinutes() - offsetMinutes);

    return date;
  } else {
    throw new Error('Invalid timestamp format');
  }
}

// 保留两位小数
export function roundToDecimalPlaces(number: number, decimalPlaces: number = 2) {
  const multiplier = Math.pow(10, decimalPlaces);
  return Math.round(number * multiplier) / multiplier;
}

export function getDevicePixelRatio(): number {
  return Math.min(window.devicePixelRatio, 2) || 1
}

function getScrollCoordinates(event: MouseEvent | TouchEvent, element: HTMLDivElement): { x: number; y: number } | null {
  const scroll: { x: number; y: number } = {
    x: 0,
    y: 0
  }
  const screenLeft = element.scrollLeft || 0;
  const screenTop = element.scrollTop || 0;

  if (event instanceof TouchEvent) {
    event = event.touches[0];
    scroll.x = event ? (screenLeft + event.clientX) : 0;
    scroll.y = screenTop + event.clientY;
  } else {
    scroll.x = screenLeft + event.pageX;
    scroll.y = screenTop + event.pageY;
  }

  return scroll; 
}
